program traffic

  implicit none

  integer :: i, iter, nmove, ncars, icell, cellval

  integer, parameter :: N = 1000000
  integer, parameter :: maxiter = 200000000/N
  integer, parameter :: printfreq = maxiter/10

  real density, rng

  integer, allocatable, dimension(:) :: cn
  integer, allocatable, dimension(:) :: cnminus1

! Set target density of cars

  density = 0.712

  write(*,*) 'Length of road is ', N
  write(*,*) 'Number of iterations is ', maxiter
  write(*,*) 'Target density of cars is ', density
  write(*,*)

! Allocate arrays

  allocate(cn(0:N+1))
  allocate(cnminus1(0:N+1))

! Initialise road accordingly using random number generator

  write(*,*) 'Initialising ...'

  ncars = 0

  do i = 1, N

! Set the value based on the global cell number for reproducibility

    call initcell(cellval, i, density)

    cnminus1(i) = cellval
    ncars = ncars + cnminus1(i)

  end do

  write(*,*) '... done'

  write(*,*) 'Actual Density of cars is ', float(ncars)/float(N)
  write(*,*)

  do iter = 1, maxiter

    nmove = 0

    cnminus1(0)   = cnminus1(N)
    cnminus1(N+1) = cnminus1(1)

! Apply CA rules to all cells

    do i = 1, N

      if (cnminus1(i) .eq. 1) then

        if (cnminus1(i+1) .eq. 1) then
          cn(i) = 1
        else
          cn(i) = 0
          nmove = nmove + 1
        end if

      else

        if (cnminus1(i-1) .eq. 1) then
          cn(i) = 1
        else
          cn(i) = 0
        end if

      end if

    end do

! Copy new to old array

    do i = 1, N
      cnminus1(i) = cn(i)
    end do

    if (mod(iter, printfreq) .eq. 0) then

        write(*,*) 'At iteration ', iter, ' average velocity is ', &
              float(nmove)/float(ncars)
    end if

 end do

 write(*,*) 'Finished'

end program traffic


subroutine initcell(cell, index, density)

  implicit none

  integer :: cell, index, i 
  real    :: density, rng

  integer (kind=8), parameter :: aa = 16807 
  integer (kind=8), parameter :: mm = 2147483647   
  integer (kind=8) :: tmp, res

  res = index*aa
  do i=1,20
     res = mod(aa*res,mm)
  end do  
  rng = dble(res)/dble(mm) 

  if (rng .lt. density) then
    cell = 1
  else
    cell = 0
  end if

end subroutine initcell 
